import networkx as nx

import pytest
from evrouting.gasstation.routing import (
    shortest_path,
    contract_graph,
    dijkstra,
    fold_path,
    get_possible_arriving_soc
)
from evrouting.graph_tools import (
    label,
    CONSUMPTION_KEY,
    DISTANCE_KEY,
    CHARGING_COEFFICIENT_KEY
)
from tests.config import edge_case, get_graph, gasstation, init_config


class TestDjikstra:
    @pytest.mark.parametrize("u,v,min_len", [(0, 1, 1), (0, 2, 3), (0, 4, 2), (0, 6, 3)])
    def test_djikstra(self, u, v, min_len):
        conf: dict = init_config(gasstation)
        shortest_path = dijkstra(conf['G'], u, v, CONSUMPTION_KEY)
        assert fold_path(conf['G'], shortest_path, CONSUMPTION_KEY) == min_len

    def test_djikstra_via_node(self):
        conf: dict = init_config(edge_case)
        shortest_path = dijkstra(conf['G'], 0, 2, CONSUMPTION_KEY)

        assert fold_path(conf['G'], shortest_path, CONSUMPTION_KEY) == 2


class TestContraction:
    def label_map(self, G: nx.Graph) -> dict:
        return {label(G, n): n for n in G.nodes}

    def test_contraction(self):
        conf: dict = init_config(gasstation)
        G: nx.Graph = conf['G']
        H: nx.Graph = contract_graph(G,
                                     conf['charging_stations'],
                                     conf['capacity'])

        # Check available gas stations
        assert set(H.nodes) == conf['charging_stations']

    @pytest.mark.parametrize('u,v,weight,value', [
        ('s', 'a', CONSUMPTION_KEY, 1),
        ('s', 'a', DISTANCE_KEY, 1),
        ('s', 'f', '', None),  # Not exist
        ('s', 'b', '', None),  # Not exist
        ('s', 'd', '', None),  # Not exist
        ('s', 'c', CONSUMPTION_KEY, 2),
        ('s', 'c', DISTANCE_KEY, 2),
        ('s', 'e', '', None)
    ])
    def test_contraction_edges(self, u, v, weight, value):
        """
        Test edges and values of its weights.
        :param u: Node
        :param v: Node
        :param weight: Weight key to check.
        :param value: Value the edge weight should have.
            If None, it tests if the egde does not exist.
        """
        conf: dict = init_config(gasstation)
        H: nx.Graph = contract_graph(conf['G'],
                                     conf['charging_stations'], conf['capacity'])

        label_map = self.label_map(H)
        try:
            edge = (label_map[u], label_map[v])
        except KeyError:
            # One of the edges is no an charging station
            if value is None:
                return True
            raise

        if value is None:
            assert edge not in H.edges
        else:
            assert H.edges[edge][weight] == value


class TestPossibleArrivingSoC:
    U = 4
    w_lower = 3
    w_greater = 5

    @pytest.fixture
    def graph_w_below_U(self):
        G = nx.Graph()
        G.add_edge(0, 1, **{CONSUMPTION_KEY: self.w_lower, DISTANCE_KEY: 2 * self.w_lower})
        yield G
        del G

    @pytest.fixture
    def graph_w_gt_U(self):
        G = nx.Graph()
        G.add_edge(0, 1, **{CONSUMPTION_KEY: self.w_greater, DISTANCE_KEY: 2 * self.w_greater})
        yield G
        del G

    @pytest.fixture
    def graph_w_eq_U(self):
        G = nx.Graph()
        G.add_edge(0, 1, **{CONSUMPTION_KEY: self.U, DISTANCE_KEY: 2 * self.U})
        yield G
        del G

    def test_unequal_charging_coeff_w_eq_U(self, graph_w_eq_U):
        G = graph_w_eq_U
        G.add_node(0, **{CHARGING_COEFFICIENT_KEY: 1})
        G.add_node(1, **{CHARGING_COEFFICIENT_KEY: 2})

        gv_lower = get_possible_arriving_soc(G, 0, self.U)
        assert gv_lower == [0]

        gv_higher = get_possible_arriving_soc(G, 1, self.U)
        assert gv_higher == [0]

    def test_equal_charging_coeff_w_lt_U(self, graph_w_below_U):
        G = graph_w_below_U
        G.add_node(0, **{CHARGING_COEFFICIENT_KEY: 1})
        G.add_node(1, **{CHARGING_COEFFICIENT_KEY: 1})

        gv_lower = get_possible_arriving_soc(G, 0, self.U)
        assert gv_lower == [0]

        gv_higher = get_possible_arriving_soc(G, 1, self.U)
        assert gv_higher == [0]

    def test_unequal_charging_coeff_w_lt_U(self, graph_w_below_U):
        G = graph_w_below_U
        G.add_node(0, **{CHARGING_COEFFICIENT_KEY: 1})
        G.add_node(1, **{CHARGING_COEFFICIENT_KEY: 2})

        gv_lower = get_possible_arriving_soc(G, 0, self.U)
        assert gv_lower == [0]

        gv_higher = get_possible_arriving_soc(G, 1, self.U)
        assert gv_higher == [0, self.U - self.w_lower]

    def test_equal_charging_coeff_w_eq_U(self, graph_w_eq_U):
        G = graph_w_eq_U
        G.add_node(0, **{CHARGING_COEFFICIENT_KEY: 1})
        G.add_node(1, **{CHARGING_COEFFICIENT_KEY: 1})

        gv_lower = get_possible_arriving_soc(G, 0, self.U)
        assert gv_lower == [0]

        gv_higher = get_possible_arriving_soc(G, 1, self.U)
        assert gv_higher == [0]

    @pytest.mark.parametrize('c_0,c_1', [(1, 2), (1, 1)])
    def test_equal_unreachable(self, graph_w_gt_U, c_0, c_1):
        G: nx.Graph = graph_w_gt_U
        G.add_node(0, **{CHARGING_COEFFICIENT_KEY: c_0})
        G.add_node(1, **{CHARGING_COEFFICIENT_KEY: c_1})

        gv_lower = get_possible_arriving_soc(G, 0, self.U)
        assert gv_lower == [0]

        gv_higher = get_possible_arriving_soc(G, 1, self.U)
        assert gv_higher == [0]


class TestRouting:
    def test_shortest_path(self):
        G = get_graph(edge_case)

        path = shortest_path(
            G,
            s=edge_case['s'],
            t=edge_case['t'],
            b_0=edge_case['b_0'],
            b_t=edge_case['b_t'],
            U=edge_case['U']
        )

        assert path == [(0, 4), (2, 0)]
